import vtk


class LiverSegmentPicker:
    def __init__(self, renderer, interactor, actors, volumes, selected_volume_actor):
        self.ren = renderer
        self.iren = interactor
        self.actors = actors
        self.actor_keys = {v: k for k, v in actors.items()}
        self.volumes = volumes
        self.selected_volume_actor = selected_volume_actor
        self.selected = set()
        self._orig = {}
        self._opacity_selected = 1.0
        self._opacity_not_selected = 0.25

        self.picker = vtk.vtkCellPicker()
        self.picker.SetTolerance(0.001)

        # Use trackball so it rotates only when you move
        self.style = vtk.vtkInteractorStyleTrackballCamera()
        self.iren.SetInteractorStyle(self.style)

        # Hook events
        self.iren.AddObserver("LeftButtonPressEvent", self.on_left_down)
        self.iren.AddObserver("LeftButtonReleaseEvent", self.on_left_up)

        self._allow_camera_drag = True

    def _save(self, a):
        if a in self._orig:
            return

    def _highlight(self, a, on):
        self._save(a)
        p = a.GetProperty()
        if on:
            p.SetOpacity(self._opacity_selected)
        else:
            p.SetOpacity(self._opacity_not_selected)

    def clear(self):
        for a in list(self.selected):
            self._highlight(a, False)
        self.selected.clear()

    def toggle(self, a):
        if a in self.selected:
            self._highlight(a, False)
            self.selected.remove(a)
        else:
            self._highlight(a, True)
            self.selected.add(a)
        total_selected_volume = 0
        for a in self.selected:
            a_key = self.actor_keys[a]
            volume = int(self.volumes[a_key])
            total_selected_volume += volume
        self.selected_volume_actor.SetInput(f'Selected volume: {total_selected_volume} mL')
            

    def on_left_down(self, iren, evt):
        x, y = iren.GetEventPosition()

        # Only do picking when Shift or Ctrl is held.
        do_pick = (iren.GetShiftKey() or iren.GetControlKey())

        if do_pick:
            self.picker.Pick(x, y, 0, self.ren)
            a = self.picker.GetActor()
            a_key = self.actor_keys[a]

            if a in self.actors.values():
                # Additive (Shift/Ctrl): toggle selection
                self.toggle(a)
            else:
                # Shift/Ctrl + empty click: clear selection (optional)
                self.clear()

            iren.GetRenderWindow().Render()
            self._allow_camera_drag = False  # don't also start a camera drag
            return

        # No modifier => normal rotate/pan/zoom interaction
        self._allow_camera_drag = True
        self.style.OnLeftButtonDown()

    def on_left_up(self, iren, evt):
        if self._allow_camera_drag:
            self.style.OnLeftButtonUp()

    def set_opacity_selected(self, opacity_selected):
        self._opacity_selected = opacity_selected

    def set_opacity_not_selected(self, opacity_not_selected):
        self._opacity_not_selected = opacity_not_selected



MOSAMATIC2_BUNDLE_IDENTIFIER = 'org.mumc'
MOSAMATIC2_APP_NAME = 'mosamatic2'
MOSAMATIC2_VERSION_FILE = 'mosamatic2/ui/resources/VERSION'
MOSAMATIC2_IMAGES_DIR_PATH = 'mosamatic2/ui/resources/images'
MOSAMATIC2_ICONS_DIR_PATH = 'mosamatic2/ui/resources/icons'
MOSAMATIC2_APP_ICON_FILE_NAME_WIN = 'mosamatic2.ico'
MOSAMATIC2_APP_ICON_FILE_NAME_MAC = 'mosamatic2.icns'
MOSAMATIC2_BACKGROUND_IMAGE_FILE_NAME = 'body-composition.jpg'
MOSAMATIC2_BACKGROUND_IMAGE_OPACITY = 0.4

MOSAMATIC2_SERVER_PORT = 8000
MOSAMATIC2_SERVER_DEBUG = True

MOSAMATIC2_WINDOW_TITLE = 'Mosamatic'
MOSAMATIC2_WINDOW_W = 1000
MOSAMATIC2_WINDOW_H = 600
MOSAMATIC2_WINDOW_GEOMETRY_KEY = 'mainwindow/geometry'
MOSAMATIC2_WINDOW_STATE_KEY = 'mainwindow/state'

MOSAMATIC2_SPLASH_SCREEN_TITLE = MOSAMATIC2_WINDOW_TITLE
MOSAMATIC2_SPLASH_SCREEN_TITLE_STYLESHEET = 'color: rgb(64, 64, 64); font-weight: bold; font-size: 32pt;'
MOSAMATIC2_SPLASH_SCREEN_SUB_TEXT = 'This software is for research only'
MOSAMATIC2_SPLASH_SCREEN_SUB_TEXT_STYLE_SHEET = 'color: rgb(64, 64, 64); font-style: italic; font-size: 10pt;'
MOSAMATIC2_SPLASH_SCREEN_START_BUTTON_TEXT = 'Start Application'
MOSAMATIC2_SPLASH_SCREEN_QUIT_BUTTON_TEXT = 'Quit'
MOSAMATIC2_SPLASH_SCREEN_W = 600
MOSAMATIC2_SPLASH_SCREEN_H = 300


import os
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.data.filedata import FileData
from mosamatic2.core.data.niftiimage import NiftiImage

LOG = LogManager()


class MultiNiftiImage(FileData):
    def __init__(self):
        super(MultiNiftiImage, self).__init__()
        self._images = []

    def images(self):
        return self._images
    
    def load(self):
        if self.path():
            for f in os.listdir(self.path()):
                f_path = os.path.join(self.path(), f)
                image = NiftiImage()
                image.set_path(f_path)
                if image.load():
                    self._images.append(image)
            return True
        return False


import shutil
from PySide6.QtWidgets import (
    QLineEdit,
    QLabel,
    QSpinBox,
    QHBoxLayout,
    QVBoxLayout,
    QFormLayout,
    QPushButton,
    QFileDialog,
)
from mosamatic2.ui.widgets.panels.visualizations.visualization import Visualization
from mosamatic2.ui.widgets.panels.visualizations.sliceselectionvisualization.sliceselectionviewer import SliceSelectionViewer
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.ui.settings import Settings
from mosamatic2.ui.utils import is_macos

LOG = LogManager()
PANEL_TITLE = 'SliceSelectionVisualization'
PANEL_NAME = 'sliceselectionvisualization'


class SliceSelectionVisualization(Visualization):
    def __init__(self):
        super(SliceSelectionVisualization, self).__init__()
        self.set_title(PANEL_TITLE)
        self._images_line_edit = None
        self._images_dir_select_button = None
        self._output_dir_line_edit = None
        self._output_dir_select_button = None
        self._load_images_button = None
        self._copy_selected_images_button = None
        self._slice_selection_viewer = None
        self._form_layout = None
        self._settings = None
        self.init_layout()

    def images_line_edit(self):
        if not self._images_line_edit:
            self._images_line_edit = QLineEdit(self.settings().get(f'{PANEL_NAME}/images'))
        return self._images_line_edit
    
    def images_dir_select_button(self):
        if not self._images_dir_select_button:
            self._images_dir_select_button = QPushButton('Select directory')
            self._images_dir_select_button.clicked.connect(self.handle_images_dir_select_button)
        return self._images_dir_select_button
    
    def output_dir_line_edit(self):
        if not self._output_dir_line_edit:
            self._output_dir_line_edit = QLineEdit(self.settings().get(f'{PANEL_NAME}/output_dir'))
        return self._output_dir_line_edit
    
    def output_dir_select_button(self):
        if not self._output_dir_select_button:
            self._output_dir_select_button = QPushButton('Select directory')
            self._output_dir_select_button.clicked.connect(self.handle_output_dir_select_button)
        return self._output_dir_select_button
    
    def copy_selected_images_button(self):
        if not self._copy_selected_images_button:
            self._copy_selected_images_button = QPushButton("Copy selected images to output directory")
            self._copy_selected_images_button.clicked.connect(self.handle_copy_selected_images_button)
        return self._copy_selected_images_button
    
    def load_images_button(self):
        if not self._load_images_button:
            self._load_images_button = QPushButton('Load')
            self._load_images_button.clicked.connect(self.handle_load_images_button)
        return self._load_images_button
    
    def slice_selection_viewer(self):
        if not self._slice_selection_viewer:
            self._slice_selection_viewer = SliceSelectionViewer()
        return self._slice_selection_viewer
    
    def form_layout(self):
        if not self._form_layout:
            self._form_layout = QFormLayout()
            if is_macos():
                self._form_layout.setFieldGrowthPolicy(QFormLayout.ExpandingFieldsGrow)
        return self._form_layout

    def settings(self):
        if not self._settings:
            self._settings = Settings()
        return self._settings

    def init_layout(self):
        image_layout = QHBoxLayout()
        image_layout.addWidget(self.images_line_edit())
        image_layout.addWidget(self.images_dir_select_button())
        output_dir_layout = QHBoxLayout()
        output_dir_layout.addWidget(self.output_dir_line_edit())
        output_dir_layout.addWidget(self.output_dir_select_button())
        self.form_layout().addRow('PNG images directory', image_layout)
        self.form_layout().addRow('Output directory', output_dir_layout)
        layout = QVBoxLayout()
        layout.addLayout(self.form_layout())
        layout.addWidget(self.load_images_button())
        layout.addWidget(self.copy_selected_images_button())
        layout.addWidget(self.slice_selection_viewer())
        self.setLayout(layout)
        self.setObjectName(PANEL_NAME)

    def handle_images_dir_select_button(self):
        last_directory = self.settings().get('last_directory')
        dir_path = QFileDialog.getExistingDirectory(dir=last_directory)
        if dir_path:
            self.images_line_edit().setText(dir_path)
            self.settings().set('last_directory', dir_path)

    def handle_output_dir_select_button(self):
        last_directory = self.settings().get('last_directory')
        dir_path = QFileDialog.getExistingDirectory(dir=last_directory)
        if dir_path:
            self.output_dir_line_edit().setText(dir_path)
            self.settings().set('last_directory', dir_path)

    def handle_load_images_button(self):
        self.slice_selection_viewer().load_images(self.images_line_edit().text())

    def handle_copy_selected_images_button(self):
        selected_paths = self.slice_selection_viewer().selected_paths()
        for p in selected_paths:
            p_dcm = p[:-13] + '.dcm'
            target_dir = self.output_dir_line_edit().text()
            LOG.info(f'Copying {p_dcm} to {target_dir}')
            shutil.copy(p_dcm, target_dir)

    def save_inputs_and_parameters(self):
        self.settings().set(f'{PANEL_NAME}/images', self.images_line_edit().text())
        self.settings().set(f'{PANEL_NAME}/output_dir', self.output_dir_line_edit().text())


import os
import csv
import numpy as np
import pandas as pd
from mosamatic2.core.tasks.task import Task
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.data.multidicomimage import MultiDicomImage
from mosamatic2.core.data.numpyimage import NumpyImage
from mosamatic2.core.utils import (
    get_pixels_from_dicom_object,
    calculate_area,
    calculate_index,
    calculate_bmi,
    calculate_sarcopenia,
    calculate_sarcopenic_obesity,
    calculate_myosteatosis,
    calculate_visceral_obesity,
    calculate_mean_radiation_attenuation,
    calculate_lama_percentage,
    get_pixels_from_tag_file,
    MUSCLE,
    SAT,
    VAT,
)

LOG = LogManager()


class CalculateScoresTask(Task):
    INPUTS = [
        'images', 
        'segmentations',
        'info',
    ]
    PARAMS = ['file_type']

    def __init__(self, inputs, params, output, overwrite=True):
        super(CalculateScoresTask, self).__init__(inputs, params, output, overwrite)

    def collect_img_seg_pairs(self, images, segmentations, file_type='npy'):
        file_type = '.tag' if file_type == 'tag' else '.seg.npy'
        img_seg_pairs = []
        for image in images:
            f_img_path = image.path()
            f_img_name = os.path.split(f_img_path)[1]
            for f_seg_path in segmentations:
                f_seg_name = os.path.split(f_seg_path)[1]
                if file_type == '.seg.npy':
                    f_seg_name = f_seg_name.removesuffix(file_type)
                    if f_seg_name == f_img_name:
                        img_seg_pairs.append((image, f_seg_path))
                elif file_type == '.tag':
                    f_seg_name = f_seg_name.removesuffix(file_type).removesuffix('.dcm')
                    f_img_name = f_img_name.removesuffix('.dcm')
                    if f_seg_name == f_img_name:
                        img_seg_pairs.append((image, f_seg_path))
                else:
                    raise RuntimeError('Unknown file type')
        return img_seg_pairs

    def load_images(self):
        image_data = MultiDicomImage()
        image_data.set_path(self.input('images'))
        if image_data.load():
            return image_data
        raise RuntimeError('Could not load images')
    
    def load_pixels_and_spacing(self, image):
        p = image.object()
        pixels = get_pixels_from_dicom_object(p, normalize=True)
        return pixels, p.PixelSpacing

    def load_segmentations(self, file_type='npy'):
        file_type = '.tag' if file_type == 'tag' else '.seg.npy'
        segmentations = []
        for f in os.listdir(self.input('segmentations')):
            f_path = os.path.join(self.input('segmentations'), f)
            if f.endswith(file_type):
                segmentations.append(f_path)
        if len(segmentations) == 0:
            raise RuntimeError('Input directory has no segmentation files')
        return segmentations

    def load_segmentation(self, f, file_type='npy'):
        if file_type == 'npy':
            segmentation = NumpyImage()
            segmentation.set_path(f)
            if segmentation.load():
                return segmentation.object()
            LOG.error(f'Could not load segmentation file {f}')
        if file_type == 'tag':
            pixels = get_pixels_from_tag_file(f)
            try:
                pixels = pixels.reshape(512, 512)
                return pixels
            except Exception:
                LOG.warning(f'Could not reshape TAG pixels to (512, 512), skipping...')
                return None
        raise RuntimeError('Unknown file type')

    def load_patient_info(self, f):
        columns = ['file', 'height', 'weight', 'sex', 'age']
        with open(f, mode='r', encoding='utf-8') as f_obj:
            reader = csv.DictReader(f_obj)
            if set(columns).issubset(reader.fieldnames):
                return [row for row in reader]
            LOG.error(f'Patient info CSV misses columns from {columns}')
        return None
        
    def get_patient_height(self, file_name, patient_info):
        for row in patient_info:
            if row['file'] in file_name:
                return float(row['height'])
        return None
    
    def get_patient_weight(self, file_name, patient_info):
        for row in patient_info:
            if row['file'] in file_name:
                return float(row['weight'])
        return None
    
    def get_patient_sex(self, file_name, patient_info) -> str:
        for row in patient_info:
            if row['file'] in file_name:
                if row['sex'] in ['male', 'female']:
                    return row['sex']
                else:
                    LOG.error(f'Sex should be either male or female in patient info CSV')
        return 'unknown'
    
    def get_patient_age(self, file_name, patient_info):
        for row in patient_info:
            if row['file'] in file_name:
                return float(row['age'])
        return None
    
    def run(self):
        image_data = self.load_images()
        images = image_data.images()
        file_type = self.param('file_type')
        segmentations = self.load_segmentations(file_type)
        img_seg_pairs = self.collect_img_seg_pairs(images, segmentations, file_type)
        patient_info = None
        patient_info_file = self.input('info')
        if patient_info_file:
            patient_info = self.load_patient_info(patient_info_file)
        # Create empty data dictionary
        data = {
            'file': [], 
            'muscle_area': [], 'muscle_idx': [], 'muscle_ra': [], 'muscle_lama_perc': [],
            'vat_area': [], 'vat_idx': [], 'vat_ra': [],
            'sat_area': [], 'sat_idx': [], 'sat_ra': [],
            'bmi': [],
            'sarcopenia': [],
            'sarcopenic_obesity': [],
            'myosteatosis': [],
            'visceral_obesity': [],
        }
        nr_steps = len(img_seg_pairs)
        for step in range(nr_steps):
            # Get image and its pixel spacing
            image, pixel_spacing = self.load_pixels_and_spacing(img_seg_pairs[step][0])
            if image is None:
                raise RuntimeError(f'Could not load DICOM image for file {img_seg_pairs[step][0]}')
            # Get segmentation for this image
            segmentation = self.load_segmentation(img_seg_pairs[step][1], file_type)
            if segmentation is None:
                LOG.warning(f'Could not load segmentation for file {img_seg_pairs[step][1]}')
                continue
            # Calculate metrics
            file_name = os.path.split(img_seg_pairs[step][0].path())[1]
            muscle_area = calculate_area(segmentation, MUSCLE, pixel_spacing)
            muscle_idx = 0
            if patient_info:
                muscle_idx = calculate_index(muscle_area, self.get_patient_height(file_name, patient_info))
            muscle_ra = calculate_mean_radiation_attenuation(image, segmentation, MUSCLE)
            muscle_lama_perc = calculate_lama_percentage(image, segmentation, MUSCLE)
            vat_area = calculate_area(segmentation, VAT, pixel_spacing)
            vat_idx = 0
            if patient_info:
                vat_idx = calculate_index(vat_area, self.get_patient_height(file_name, patient_info))
            vat_ra = calculate_mean_radiation_attenuation(image, segmentation, VAT)
            sat_area = calculate_area(segmentation, SAT, pixel_spacing)
            sat_idx = 0
            if patient_info:
                sat_idx = calculate_index(sat_area, self.get_patient_height(file_name, patient_info))
            sat_ra = calculate_mean_radiation_attenuation(image, segmentation, SAT)
            sex = 'unknown'
            if patient_info:
                sex = self.get_patient_sex(file_name, patient_info)
            bmi = 0
            if patient_info:
                bmi = calculate_bmi(
                    self.get_patient_weight(file_name, patient_info), 
                    self.get_patient_height(file_name, patient_info),
                )
            sarcopenia = 'no'
            if patient_info:
                sarcopenia = calculate_sarcopenia(muscle_idx, bmi, sex)
            sarcopenic_obesity = 'no'
            if patient_info:
                sarcopenic_obesity = calculate_sarcopenic_obesity(muscle_idx, bmi, sex)
            myosteatosis = 'no'
            if patient_info:
                myosteatosis = calculate_myosteatosis(muscle_ra, bmi)
            visceral_obesity = 'no'
            if patient_info:
                visceral_obesity = calculate_visceral_obesity(vat_area)
            LOG.info(f'file: {file_name}, ' +
                    f'muscle_area: {muscle_area}, muscle_idx: {muscle_idx}, muscle_ra: {muscle_ra}, ' +
                    f'vat_area: {vat_area}, vat_idx: {vat_idx}, vat_ra: {vat_ra}, ' +
                    f'sat_area: {sat_area}, sat_idx: {sat_idx}, sat_ra: {sat_ra}, ' +
                    f'muscle_lama_perc: {muscle_lama_perc}, ' +
                    f'bmi: {bmi}, ' +
                    f'sarcopenia: {sarcopenia}, ' +
                    f'sarcopenic_obesity: {sarcopenic_obesity}, ' +
                    f'myosteatosis: {myosteatosis}, ' +
                    f'visceral_obesity: {visceral_obesity}'
            )
            # Update dataframe data
            data['file'].append(file_name)
            data['muscle_area'].append(muscle_area)
            data['muscle_idx'].append(muscle_idx)
            data['muscle_ra'].append(muscle_ra)
            data['muscle_lama_perc'].append(muscle_lama_perc)
            data['vat_area'].append(vat_area)
            data['vat_idx'].append(vat_idx)
            data['vat_ra'].append(vat_ra)
            data['sat_area'].append(sat_area)
            data['sat_idx'].append(sat_idx)
            data['sat_ra'].append(sat_ra)
            data['bmi'].append(bmi)
            data['sarcopenia'].append(sarcopenia)
            data['sarcopenic_obesity'].append(sarcopenic_obesity)
            data['myosteatosis'].append(myosteatosis)
            data['visceral_obesity'].append(visceral_obesity)
            # Update progress
            self.set_progress(step, nr_steps)
        # Build dataframe and return the CSV file as output
        csv_file_path = os.path.join(self.output(), 'bc_scores.csv')
        xls_file_path = os.path.join(self.output(), 'bc_scores.xlsx')
        df = pd.DataFrame(data=data)
        df.to_csv(csv_file_path, index=False, sep=';')
        df.to_excel(xls_file_path, index=False, engine='openpyxl')


import os
import pandas as pd
from mosamatic2.core.tasks.task import Task
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.utils import (
    is_dicom, 
    load_dicom,
    current_time_in_seconds,
    elapsed_time_in_seconds,
    duration,
)

LOG = LogManager()


class CreateDicomSummaryTask(Task):
    INPUTS = ['directory']
    PARAMS = []

    def __init__(self, inputs, params, output, overwrite):
        super(CreateDicomSummaryTask, self).__init__(inputs, params, output, overwrite)

    def get_values(self, files, tag_name):
        values = []
        for f in files:
            if tag_name in f.keys():
                if f[tag_name] not in values:
                    values.append(f[tag_name])
        return values
    
    def get_tag_value(self, p, tag_name):
        return p.data_element(tag_name).value
    
    def run(self):
        series = {}
        start_time = current_time_in_seconds()
        for root, dirs, files in os.walk(self.input('directory')):
            for f in files:
                f_path = os.path.join(root, f)
                if is_dicom(f_path):
                    p = load_dicom(f_path, stop_before_pixels=True)
                    if p:
                        patient_id = self.get_tag_value(p, 'PatientID')
                        if patient_id:
                            if patient_id not in series.keys():
                                series[patient_id] = {}
                            series_instance_uid = self.get_tag_value(p, 'SeriesInstanceUID')
                            if series_instance_uid:
                                if series_instance_uid not in series[patient_id].keys():
                                    series[patient_id][series_instance_uid] = []
                                series[patient_id][series_instance_uid].append({
                                    'series_description': self.get_tag_value(p, 'SeriesDescription'),
                                    'slice_thickness': self.get_tag_value(p, 'SliceThickness'),
                                    'rows': self.get_tag_value(p, 'Rows'),
                                    'columns': self.get_tag_value(p, 'Columns'),
                                    'pixel_spacing': self.get_tag_value(p, 'PixelSpacing'),
                                    'modality': self.get_tag_value(p, 'Modality'),
                                    'image_type': self.get_tag_value(p, 'ImageType'),
                                })
        data = {
            'patient_id': [],
            'series_description': [],
            'nr_images': [],
            'slice_thickness': [],
            'rows': [],
            'columns': [],
            'pixel_spacing': [],
            'modality': [],
            'image_type': [],
            'series_instance_uid': [],
        }
        for patient_id, v1 in series.items():
            for series_instance_uid, v2 in v1.items():
                data['patient_id'].append(patient_id)
                data['series_description'].append(self.get_values(v2, tag_name='series_description'))
                data['nr_images'].append(len(v2))
                data['slice_thickness'].append(self.get_values(v2, tag_name='slice_thickness'))
                data['rows'].append(self.get_values(v2, tag_name='rows'))
                data['columns'].append(self.get_values(v2, tag_name='columns'))
                data['pixel_spacing'].append(self.get_values(v2, tag_name='pixel_spacing'))
                data['modality'].append(self.get_values(v2, tag_name='modality'))
                data['image_type'].append(self.get_values(v2, tag_name='image_type'))
                data['series_instance_uid'].append(series_instance_uid)
        df = pd.DataFrame(data=data)
        df.to_csv(os.path.join(self.output(), 'summary.csv'), index=False, sep=';')
        df.to_excel(os.path.join(self.output(), 'summary.xlsx'), index=False, engine='openpyxl')
        LOG.info(df)
        LOG.info(f'Elapsed time: {duration(elapsed_time_in_seconds(start_time))}')


import click

from mosamatic2.core.pipelines import LiverAnalysisPipeline


@click.command(help='Runs liver analysis pipeline')
@click.option(
    '--scans', 
    required=True, 
    type=click.Path(exists=True), 
    help='Root directory with scan directories for each patient (no spaces allowed)',
)
@click.option(
    '--compressed', 
    default=True,
    help='Whether to produce compressed NIFTI file or not (default: True)'
)
@click.option(
    '--output', 
    required=True, 
    type=click.Path(), 
    help='Output directory (no spaces allowed)'
)
@click.option(
    '--overwrite', 
    type=click.BOOL, 
    default=False, 
    help='Overwrite [true|false]'
)
def liveranalysispipeline(scans, compressed, output, overwrite):
    """
    Runs liver analysis pipeline. This pipeline run the following steps on each scan: 

    (1) DICOM to NIFTI conversion
    (2) Extract liver segments using Total Segmentator
    (3) Calculate segment statistics, e.g., volume (mL), mean HU, std HU and PNG image
        of each segments HU histogram.
    
    Parameters
    ----------
    --scans : str
        Root directory with scan directories for each patient. Each scan directory should 
        contain DICOM images for a single scan only and nothing else. Also, the directory
        paths cannot contain any spaces.

    --compressed : str
        Whether to produce compressed NIFTI file or not (default: True)

    --output : str
        Path to output directory. No spaces allowed.

    --overwrite : bool
        Overwrite contents output directory [true|false]
    """
    pipeline = LiverAnalysisPipeline(
        inputs={'scans': scans},
        params={'compressed': compressed},
        output=output,
        overwrite=overwrite,
    )
    pipeline.run()


from mosamatic2.core.data.filedata import FileData
from mosamatic2.core.data.dicomimage import DicomImage
from mosamatic2.core.data.dicomimageseries import DicomImageSeries
from mosamatic2.core.data.multidicomimage import MultiDicomImage
from mosamatic2.core.data.dixonseries import DixonSeries


import os
from mosamatic2.core.data.filedata import FileData
from mosamatic2.core.data.dicomimage import DicomImage


class DicomImageSeries(FileData):
    def __init__(self):
        super(DicomImageSeries, self).__init__()
        self._images = []

    def images(self):
        return self._images
    
    def load(self):
        if self.path():
            images = []
            for f in os.listdir(self.path()):
                f_path = os.path.join(self.path(), f)
                image = DicomImage()
                image.set_path(f_path)
                if image.load():
                    images.append(image)
            # Sort DICOM objects by instance number
            self._images = sorted(images, key=lambda image: int(image.object().get('InstanceNumber')))
            return True
        return False


import os
import atexit
import datetime
from pathlib import Path
from mosamatic2.core.singleton import singleton


@singleton
class LogManager:
    def __init__(self, suppress_print=False):
        self._suppress_print = suppress_print
        self._listeners = []
        file_path = os.path.join(Path.home(), 'mosamatic2.log')
        self._file_handle = open(file_path, 'w', buffering=1)
        atexit.register(self.close_file)

    def _log(self, level, message):
        timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        message = f'[{timestamp}] {level} : {message}'
        if not self._suppress_print:
            print(message)
        self._file_handle.write(message + '\n')
        self.notify_listeners(message)
        return message

    def info(self, message):
        return self._log('INFO', message)

    def warning(self, message):
        return self._log('WARNING', message)

    def error(self, message):
        return self._log('ERROR', message)
    
    def add_listener(self, listener):
        if listener not in self._listeners:
            self._listeners.append(listener)

    def notify_listeners(self, message):
        for listener in self._listeners:
            listener.new_message(message)

    def close_file(self):
        if self._file_handle:
            self._file_handle.close()


from mosamatic2.core.tasks.task import Task


class Pipeline(Task):
    def __init__(self, inputs, params, output, overwrite=False):
        super(Pipeline, self).__init__(inputs, params, output, overwrite)
        self._tasks = []

    def add_task(self, task):
        self._tasks.append(task)

    def run(self):
        for task in self._tasks:
            task.run()


import time
import os
import click
import textwrap
import math
import pendulum
import numpy as np
import nibabel as nb
import struct
import binascii
import pydicom
import warnings
from pathlib import Path
from pydicom.uid import (
    ExplicitVRLittleEndian, ImplicitVRLittleEndian, ExplicitVRBigEndian
)
from PIL import Image
from mosamatic2.core.managers.logmanager import LogManager

warnings.filterwarnings("ignore", message="Invalid value for VR UI:", category=UserWarning)

MUSCLE, VAT, SAT = 1, 5, 7
LOG = LogManager()


def create_name_with_timestamp(prefix: str='') -> str:
    tz = pendulum.local_timezone()
    timestamp = pendulum.now(tz).strftime('%Y%m%d%H%M%S%f')[:17]
    if prefix != '' and not prefix.endswith('-'):
        prefix = prefix + '-'
    name = f'{prefix}{timestamp}'
    return name


def show_doc_command(cli_group: click.Group) -> click.Command:
    @click.command(name="showdoc")
    @click.argument("command_name", required=False)
    def show_doc(command_name):
        commands = cli_group.commands
        if command_name:
            cmd = commands.get(command_name)
            if cmd and hasattr(cmd, 'callback') and cmd.callback.__doc__:
                print()
                print(textwrap.dedent(cmd.callback.__doc__).strip())
            else:
                click.echo(f'No docstring found for command: {command_name}')
        else:
            click.echo('Available commands with docstrings:')
            for name, cmd in commands.items():
                if hasattr(cmd, 'callback') and cmd.callback.__doc__:
                    click.echo(f"  {name}")
            click.echo('\nUse: `mosamatic show-doc <command>` to view a commands docstring')
    return show_doc


def home_dir():
    return Path.home()


def mosamatic_dir():
    d = os.path.join(home_dir(), '.mosamatic2')
    os.makedirs(d, exist_ok=True)
    return d


def mosamatic_data_dir():
    data_dir = os.path.join(mosamatic_dir(), 'data')
    os.makedirs(data_dir, exist_ok=True)
    return data_dir


def mosamatic_output_dir():
    output_dir = os.path.join(mosamatic_data_dir(), 'output')
    os.makedirs(output_dir, exist_ok=True)
    return output_dir


def current_time_in_milliseconds():
    return int(round(time.time() * 1000))


def current_time_in_seconds() -> int:
    return int(round(current_time_in_milliseconds() / 1000.0))


def elapsed_time_in_milliseconds(start_time_in_milliseconds):
    return current_time_in_milliseconds() - start_time_in_milliseconds


def elapsed_time_in_seconds(start_time_in_seconds):
    return current_time_in_seconds() - start_time_in_seconds


def duration(seconds):
    h = int(math.floor(seconds/3600.0))
    remainder = seconds - h * 3600
    m = int(math.floor(remainder/60.0))
    remainder = remainder - m * 60
    s = int(math.floor(remainder))
    return '{} hours, {} minutes, {} seconds'.format(h, m, s)


def is_dicom(f):
    try:
        pydicom.dcmread(f, stop_before_pixels=True)
        return True
    except pydicom.errors.InvalidDicomError:
        pass
    return False
    

def load_dicom(f, stop_before_pixels=False):
    try:
        return pydicom.dcmread(f, stop_before_pixels=stop_before_pixels)
    except pydicom.errors.InvalidDicomError:
        try:
            p = pydicom.dcmread(f, stop_before_pixels=stop_before_pixels, force=True)
            if hasattr(p, 'SOPClassUID'):
                if not hasattr(p.file_meta, 'TransferSyntaxUID'):
                    LOG.warning(f'DICOM file {f} does not have FileMetaData/TransferSyntaxUID, trying to fix...')
                    p.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
                return p
        except pydicom.errors.InvalidDicomError:
            pass
    return None


def is_jpeg2000_compressed(p):
    if hasattr(p.file_meta, 'TransferSyntaxUID'):
        return p.file_meta.TransferSyntaxUID not in [ExplicitVRLittleEndian, ImplicitVRLittleEndian, ExplicitVRBigEndian]
    return False


def is_nifti(f):
    return f.endswith('.nii') or f.endswith('.nii.gz')


def load_nifti(f):
    if is_nifti(f):
        return nb.load(f)
    return None


def is_numpy_array(value):
    return isinstance(value, np.array)


def is_numpy(f):
    try:
        np.load(f)
        return True
    except:
        return False


def load_numpy_array(f):
    if is_numpy(f):
        return np.load(f)
    return None


def get_pixels_from_tag_file(tag_file_path):
    f = open(tag_file_path, 'rb')
    f.seek(0)
    byte = f.read(1)
    # Make sure to check the byte-value in Python 3!!
    while byte != b'':
        byte_hex = binascii.hexlify(byte)
        if byte_hex == b'0c':
            break
        byte = f.read(1)
    values = []
    f.read(1)
    while byte != b'':
        v = struct.unpack('b', byte)
        values.append(v)
        byte = f.read(1)
    values = np.asarray(values)
    values = values.astype(np.uint16)
    return values


def get_rescale_params(p):
    rescale_slope = getattr(p, 'RescaleSlope', None)
    rescale_intercept = getattr(p, 'RescaleIntercept', None)
    if rescale_slope is not None and rescale_intercept is not None:
        return rescale_slope, rescale_intercept
    # Try Enhanced DICOM structure
    if 'SharedFunctionalGroupsSequence' in p:
        fg = p.SharedFunctionalGroupsSequence[0]
        if 'PixelValueTransformationSequence' in fg:
            pvt = fg.PixelValueTransformationSequence[0]
            rescale_slope = pvt.get('RescaleSlope', 1)
            rescale_intercept = pvt.get('RescaleIntercept', 0)
            return rescale_slope, rescale_intercept
    return 1, 0


def get_pixels_from_dicom_object(p, normalize=True):
    pixels = p.pixel_array
    if not normalize:
        return pixels
    if normalize is True: # Map pixel values back to original HU values
        rescale_slope, rescale_intercept = get_rescale_params(p)
        return rescale_slope * pixels + rescale_intercept
    if isinstance(normalize, int):
        return (pixels + np.min(pixels)) / (np.max(pixels) - np.min(pixels)) * normalize
    if isinstance(normalize, list):
        return (pixels + np.min(pixels)) / (np.max(pixels) - np.min(pixels)) * normalize[1] + normalize[0]
    return pixels


def convert_labels_to_157(label_image: np.array) -> np.array:
    label_image157 = np.copy(label_image)
    label_image157[label_image157 == 1] = 1
    label_image157[label_image157 == 2] = 5
    label_image157[label_image157 == 3] = 7
    return label_image157


def normalize_between(img: np.array, min_bound: int, max_bound: int) -> np.array:
    img = (img - min_bound) / (max_bound - min_bound)
    # img[img > 1] = 1
    img[img < 0] = 0
    img[img > 1] = 0
    c = (img - np.min(img))
    d = (np.max(img) - np.min(img))
    img = np.divide(c, d, np.zeros_like(c), where=d != 0)
    return img


def apply_window_center_and_width(image: np.array, center: int, width: int) -> np.array:
    image_min = center - width // 2
    image_max = center + width // 2
    windowed_image = np.clip(image, image_min, image_max)
    windowed_image = ((windowed_image - image_min) / (image_max - image_min)) * 255.0
    return windowed_image.astype(np.uint8)


def calculate_area(labels: np.array, label, pixel_spacing) -> float:
    mask = np.copy(labels)
    mask[mask != label] = 0
    mask[mask == label] = 1
    area = np.sum(mask) * (pixel_spacing[0] * pixel_spacing[1]) / 100.0
    return area


def calculate_index(area: float, height: float) -> float:
    return area / (height * height)


def calculate_bmi(weight: float, height: float) -> float:
    return weight / (height * height)


def calculate_sarcopenia(muscle_idx: float, bmi: float, sex: str) -> str:
    if not sex in ['male', 'female']:
        return 'unknown'
    if bmi >= 25 and muscle_idx < 53:
        return 'yes'
    if bmi < 25 and ((sex == 'male' and muscle_idx < 43) or (sex == 'female' and muscle_idx < 41)):
        return 'yes'
    return 'no'

def calculate_sarcopenic_obesity(muscle_idx: float, bmi: float, sex: str) -> str:
    if sex not in ['male', 'female']:
        return 'unknown'
    if bmi >= 25 and ((muscle_idx < 38.5 and sex == 'male') or (muscle_idx < 52.4 and sex == 'female')):
        return 'yes'
    return 'no'

def calculate_myosteatosis(smra: float, bmi: float) -> str:
    if (smra < 33 and bmi >= 25) or (smra < 41 and bmi < 25):
        return 'yes'
    return 'no'

def calculate_visceral_obesity(vat_area: float) -> str:
    if vat_area > 100:
        return 'yes'
    return 'no'

def calculate_mean_radiation_attenuation(image: np.array, labels: np.array, label: int) -> float:
    mask = np.copy(labels)
    mask[mask != label] = 0
    mask[mask == label] = 1
    subtracted = image * mask
    mask_sum = np.sum(mask)
    if mask_sum > 0.0:
        mean_radiation_attenuation = np.sum(subtracted) / np.sum(mask)
    else:
        # print('Sum of mask pixels is zero, return zero radiation attenuation')
        mean_radiation_attenuation = 0.0
    return mean_radiation_attenuation


def calculate_lama_percentage(image: np.ndarray, labels: np.ndarray, label: int, threshold: float = 30.0) -> float:
    roi = (labels == label)
    n_roi = int(np.count_nonzero(roi))
    if n_roi == 0:
        return 0.0
    lama = roi & (image < threshold)
    lama_pct = (np.count_nonzero(lama) / n_roi) * 100.0
    return int(lama_pct)


def calculate_dice_score(ground_truth: np.array, prediction: np.array, label: int) -> float:
    numerator = prediction[ground_truth == label]
    numerator[numerator != label] = 0
    n = ground_truth[prediction == label]
    n[n != label] = 0
    if np.sum(numerator) != np.sum(n):
        raise RuntimeError('Mismatch in Dice score calculation!')
    denominator = (np.sum(prediction[prediction == label]) + np.sum(ground_truth[ground_truth == label]))
    dice_score = np.sum(numerator) * 2.0 / denominator
    return dice_score


def convert_dicom_to_numpy_array(dicom_file_path: str, window_level: int=50, window_width: int=400, normalize=True) -> np.array:
    p = pydicom.dcmread(dicom_file_path)
    pixels = p.pixel_array
    pixels = pixels.reshape(p.Rows, p.Columns)
    if normalize:
        b = p.RescaleIntercept
        m = p.RescaleSlope
        pixels = m * pixels + b
    pixels = apply_window_center_and_width(pixels, window_level, window_width)
    return pixels


def convert_dicom_to_png_image(dicom_file_path: str, output_dir_path: str, window_level: int=50, window_width: int=400, normalize=True) -> str:
    array = convert_dicom_to_numpy_array(dicom_file_path, window_level, window_width, normalize)
    convert_numpy_array_to_png_image(
        array,
        output_dir_path,
        None,
        os.path.split(dicom_file_path)[1] + '.png',
        10, 10,
    )


class ColorMap:
    def __init__(self, name: str) -> None:
        self._name = name
        self._values = []

    def name(self) -> str:
        return self._name
    
    def values(self):
        return self._values
    

class GrayScaleColorMap(ColorMap):
    def __init__(self) -> None:
        super(GrayScaleColorMap, self).__init__(name='GrayScaleColorMap')
        # Implement your own gray scale map or let NumPy do this more efficiently?
        pass    

class AlbertaColorMap(ColorMap):
    def __init__(self) -> None:
        super(AlbertaColorMap, self).__init__(name='AlbertaColorMap')
        for i in range(256):
            if i == 1:  # muscle
                self.values().append([255, 0, 0])
            elif i == 2:  # inter-muscular adipose tissue
                self.values().append([0, 255, 0])
            elif i == 5:  # visceral adipose tissue
                self.values().append([255, 255, 0])
            elif i == 7:  # subcutaneous adipose tissue
                self.values().append([0, 255, 255])
            elif i == 12:  # unknown
                self.values().append([0, 0, 255])
            else:
                self.values().append([0, 0, 0])


def apply_color_map(pixels: np.array, color_map: ColorMap) -> np.array:
    pixels_new = np.zeros((*pixels.shape, 3), dtype=np.uint8)
    np.take(color_map.values(), pixels, axis=0, out=pixels_new)
    return pixels_new


def convert_numpy_array_to_png_image(
        numpy_array_file_path_or_object: str, output_dir_path: str, color_map: ColorMap=None, png_file_name: str=None, fig_width: int=10, fig_height: int=10) -> str:
    if isinstance(numpy_array_file_path_or_object, str):
        numpy_array = np.load(numpy_array_file_path_or_object)
    else:
        numpy_array = numpy_array_file_path_or_object
        if not png_file_name:
            raise RuntimeError('PNG file name required for NumPy array object')
    if color_map:
        numpy_array = apply_color_map(pixels=numpy_array, color_map=color_map)
    image = Image.fromarray(numpy_array)
    if not png_file_name:
        numpy_array_file_name = os.path.split(numpy_array_file_path_or_object)[1]
        png_file_name = numpy_array_file_name + '.png'      
    elif not png_file_name.endswith('.png'):
        png_file_name += '.png'
    png_file_path = os.path.join(output_dir_path, png_file_name)
    image.save(png_file_path)
    return png_file_path


def convert_muscle_mask_to_myosteatosis_map(hu: np.array, mask: np.array, output_dir: str, png_file_name: str, hu_low: int = 30, hu_high: int = 200, alpha: float = 1.0) -> str:
    muscle_mask = (mask == 1)
    red = muscle_mask & (hu >= hu_low) & (hu <= hu_high)
    yellow = muscle_mask & (hu < hu_low)
    overlay = np.zeros((*hu.shape, 4), dtype=np.float32)
    overlay[..., 3] = 1.0
    overlay[red] = (1.0, 0.0, 0.0, alpha)
    overlay[yellow] = (1.0, 1.0, 0.0, alpha)
    overlay_u8 = (np.clip(overlay, 0.0, 1.0) * 255).astype(np.uint8)
    png_file_path = os.path.join(output_dir, png_file_name)
    # image = Image.fromarray(overlay)
    image = Image.fromarray(overlay_u8, mode="RGBA")
    image.save(png_file_path)
    return overlay


def is_docker_running():
    import docker
    try:
        client = docker.from_env()
        client.ping()
        return True
    except Exception:
        return False
    

def is_path_docker_compatible(path):
    return not ' ' in path


def to_unix_path(path):
    return path.replace("\\", "/").replace(" ", "\\ ")


import mosamatic2.constants as constants
from PySide6.QtCore import QSettings
from mosamatic2.core.managers.logmanager import LogManager

LOG = LogManager()


class Settings(QSettings):
    def __init__(self):
        super(Settings, self).__init__(
            QSettings.IniFormat, 
            QSettings.UserScope, 
            constants.MOSAMATIC2_BUNDLE_IDENTIFIER, 
            constants.MOSAMATIC2_APP_NAME,
        )

    def prepend_bundle_identifier_and_name(self, name):
        return '{}.{}.{}'.format(constants.MOSAMATIC2_BUNDLE_IDENTIFIER, constants.MOSAMATIC2_APP_NAME, name)

    def get(self, name, default=None):
        if not name.startswith(constants.MOSAMATIC2_BUNDLE_IDENTIFIER):
            name = self.prepend_bundle_identifier_and_name(name)
        value = self.value(name)
        if value is None or value == '':
            return default
        return value
    
    def get_int(self, name, default=None):
        try:
            return int(self.get(name, default))
        except ValueError as e:
            return default
        
    def get_float(self, name, default=None):
        try:
            return float(self.get(name, default))
        except ValueError as e:
            return default
        
    def get_bool(self, name, default=None):
        try:
            value = self.get(name, default)
            if value and isinstance(value, str):
                if value == '0' or value.lower() == 'false':
                    return False
                elif value == '1' or value.lower() == 'true':
                    return True
                else:
                    return default
            if value and isinstance(value, bool):
                return value
        except ValueError as e:
            return default
    
    def set(self, name, value):
        name = self.prepend_bundle_identifier_and_name(name)
        self.setValue(name, value)

    def print(self):
        LOG.info(f'Settings path: {self.fileName()}')
        for key in self.allKeys():
            LOG.info(f'Settings: {key}')


import os
import torch
import numpy as np
import matplotlib.pyplot as plt

import models

from mosamatic2.core.tasks.task import Task
from mosamatic2.core.utils import (
    normalize_between,
    get_pixels_from_dicom_object,
    convert_labels_to_157,
)
from mosamatic2.core.data.multidicomimage import MultiDicomImage
from mosamatic2.core.data.dicomimage import DicomImage
from mosamatic2.core.tasks.segmentmusclefatt4pytorchtask.paramloader import ParamLoader

DEVICE = 'cpu'


class SegmentMuscleFatT4PyTorchTask(Task):
    INPUTS = [
        'images', 
        'model_files'
    ]
    PARAMS = ['model_version']

    def __init__(self, inputs, params, output, overwrite=True):
        super(SegmentMuscleFatT4PyTorchTask, self).__init__(inputs, params, output, overwrite)

    def load_images(self):        
        image_data = MultiDicomImage()
        image_data.set_path(self.input('images'))
        if image_data.load():
            return image_data
        raise RuntimeError('Could not load images')

    def load_model_files(self):
        model_files = []
        for f in os.listdir(self.input('model_files')):
            f_path = os.path.join(self.input('model_files'), f)
            if f_path.endswith('.pt') or f_path.endswith('.json'):
                model_files.append(f_path)
        if len(model_files) != 3:
            raise RuntimeError(f'Found {len(model_files)} model files. This should be 3!')
        return model_files

    def load_models_and_params(self, model_files, model_version):
        # First load params.json because it is needed to instantiate the models
        params = None
        for f_path in model_files:
            f_name = os.path.split(f_path)[1]
            if f_name == f'params-{str(model_version)}.json':
                params = ParamLoader(f_path)
                break
        if params is None:
            raise RuntimeError('Could not load parameters')
        # Next, load models
        model, contour_model = None, None
        for f_path in model_files:
            f_name = os.path.split(f_path)[1]
            if f_name == f'model-{str(model_version)}.pt':
                # model = models.AttentionUNet(params, 4).to(device=DEVICE)
                model = models.UNet(params, 4).to(device=DEVICE)
                model.load_state_dict(torch.load(f_path, weights_only=False, map_location=torch.device(DEVICE)))
                model.eval()
            elif f_name == f'contour_model-{str(model_version)}.pt':
                contour_model = models.UNet(params, 2).to(device=DEVICE)
                contour_model.load_state_dict(torch.load(f_path, weights_only=False, map_location=torch.device(DEVICE)))
                contour_model.eval()
            else:
                pass
        return model, contour_model, params
    
    def extract_contour(self, image, contour_model):
        with torch.no_grad():
            # Create 4D Tensor input
            input = np.expand_dims(image, 0)
            input = np.expand_dims(input, 0)
            input = torch.Tensor(input)
            input = input.to(DEVICE, dtype=torch.float)
            # Predict
            prediction = contour_model(input)
            prediction = torch.argmax(prediction, axis=1)
            prediction = prediction.squeeze()
            prediction = prediction.detach().cpu().numpy()
            result = image * prediction
        return result
    
    def show_result(self, result):
        fig = plt.figure()
        plt.imshow(result, cmap='gray')
        plt.savefig("/Users/ralph/result.png")
    
    def segment_muscle_and_fat(self, image, model):
        input = np.expand_dims(image, 0)
        input = np.expand_dims(input, 0)
        input = torch.Tensor(input)
        input = input.to(DEVICE, dtype=torch.float)
        segmentation = model(input)
        segmentation = torch.argmax(segmentation, axis=1)
        segmentation = segmentation.squeeze()
        segmentation = segmentation.detach().cpu().numpy()
        return segmentation
    
    def process_file(self, image, output_dir, model, contour_model, params):
        assert isinstance(image, DicomImage)
        pixels = get_pixels_from_dicom_object(image.object(), normalize=True)
        if contour_model:
            pixels = normalize_between(pixels, params.dict['lower_bound'], params.dict['upper_bound'])
            pixels = self.extract_contour(pixels, contour_model)
        pixels = pixels.astype(np.float32)
        segmentation = self.segment_muscle_and_fat(pixels, model)
        segmentation = convert_labels_to_157(segmentation)
        segmentation_file_name = os.path.split(image.path())[1]
        segmentation_file_path = os.path.join(output_dir, f'{segmentation_file_name}.seg.npy')
        np.save(segmentation_file_path, segmentation)

    def run(self):
        image_data = self.load_images()
        model_files = self.load_model_files()
        model_version = self.param('model_version')
        model, contour_model, params = self.load_models_and_params(model_files, model_version)
        images = image_data.images()
        nr_steps = len(images)
        for step in range(nr_steps):
            self.process_file(images[step], self.output(), model, contour_model, params)
            self.set_progress(step, nr_steps)


import os
import vtk
import pandas as pd
from PySide6.QtWidgets import (
    QWidget, 
    QVBoxLayout,
)
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
from mosamatic2.ui.widgets.panels.visualizations.liversegmentvisualization.liversegmentpicker import LiverSegmentPicker
from mosamatic2.core.managers.logmanager import LogManager

LOG = LogManager()
SEGMENT_COLORS = [
    (0.80, 0.25, 0.25),  # muted red
    (0.25, 0.55, 0.80),  # steel blue
    (0.30, 0.70, 0.45),  # green
    (0.85, 0.65, 0.25),  # amber
    (0.55, 0.40, 0.75),  # purple
    (0.90, 0.45, 0.15),  # orange
    (0.20, 0.75, 0.75),  # teal
    (0.75, 0.75, 0.30),  # olive
]

vtk.vtkObject.GlobalWarningDisplayOff()


class LiverSegmentViewer(QWidget):
    def __init__(self):
        super(LiverSegmentViewer, self).__init__()
        self._total_volume = 0.0
        self._selected_volume = 0.0
        self._liver_segment_actors = {}
        self._liver_volumes = {}
        self._vtk_widget = QVTKRenderWindowInteractor(self)
        self._render_window = self._vtk_widget.GetRenderWindow()
        self._interactor = self._render_window.GetInteractor()
        self._interactor.SetInteractorStyle(vtk.vtkInteractorStyleTrackballCamera())
        layout = QVBoxLayout()
        layout.addWidget(self._vtk_widget)
        self.setLayout(layout)
        self._selected_volume_actor = vtk.vtkTextActor()
        self._selected_volume_actor.SetInput(f'Selected volume: {self._selected_volume} mL')
        self._selected_volume_actor.GetTextProperty().SetFontSize(18)
        self._selected_volume_actor.GetTextProperty().SetColor(1, 1, 1)  # RGB 0–1
        self._selected_volume_actor.SetDisplayPosition(20, 50)
        self._total_volume_actor = vtk.vtkTextActor()
        self._total_volume_actor.SetInput(f'Total volume: {self._total_volume} mL')
        self._total_volume_actor.GetTextProperty().SetFontSize(18)
        self._total_volume_actor.GetTextProperty().SetColor(1, 1, 1)  # RGB 0–1
        self._total_volume_actor.SetDisplayPosition(20, 20)
        self._renderer = vtk.vtkRenderer()
        self._renderer.AddViewProp(self._total_volume_actor)
        self._renderer.AddViewProp(self._selected_volume_actor)
        self._render_window.AddRenderer(self._renderer)
        self._render_window.Render()

    def image_to_surface(self, image: vtk.vtkImageData, label=1) -> vtk.vtkPolyData:
        dmc = vtk.vtkDiscreteMarchingCubes()
        dmc.SetInputData(image)
        dmc.SetValue(0, label)
        dmc.Update()
        cleaner = vtk.vtkCleanPolyData()
        cleaner.SetInputConnection(dmc.GetOutputPort())
        cleaner.Update()
        smoother = vtk.vtkWindowedSincPolyDataFilter()
        smoother.SetInputConnection(cleaner.GetOutputPort())
        smoother.SetNumberOfIterations(20)   # 10–30 typical
        smoother.SetPassBand(0.1)              # smaller = smoother (e.g. 0.05–0.2)
        smoother.SetFeatureAngle(120.0)      # preserve sharper edges
        smoother.BoundarySmoothingOff()              # usually best for closed organs
        smoother.FeatureEdgeSmoothingOff()           # keep features; turn On if you want more smoothing
        smoother.NonManifoldSmoothingOn()
        smoother.NormalizeCoordinatesOn()            # helps numerical stability
        smoother.Update()
        normals = vtk.vtkPolyDataNormals()
        normals.SetInputConnection(smoother.GetOutputPort())
        normals.SetComputePointNormals(True)
        normals.SetComputeCellNormals(False)
        normals.SetAutoOrientNormals(True)
        normals.SetConsistency(True)
        normals.Update()
        return normals.GetOutput()

    def surface_to_actor(self, polydata: vtk.vtkPolyData) -> vtk.vtkActor:
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputData(polydata)
        mapper.ScalarVisibilityOff()
        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        return actor
    
    def load_segments_and_volumes(self, liver_segments_dir, liver_volumes_file):
        # Load volumes
        df = pd.read_csv(liver_volumes_file, sep=';')
        for _, row in df.iterrows():
            self._liver_volumes[row['file']] = row['volume_mL']
        self._total_volume = sum(self._liver_volumes.values())
        self._total_volume_actor.SetInput(f'Total volume: {self._total_volume} mL')
        # Load liver segment actors
        i = 0
        for f in os.listdir(liver_segments_dir):
            segment_name = f
            f_path = os.path.join(liver_segments_dir, f)
            reader = vtk.vtkNIFTIImageReader()
            reader.SetFileName(f_path)
            reader.Update()
            image = reader.GetOutput()
            surface = self.image_to_surface(image)
            actor = self.surface_to_actor(surface)
            prop = actor.GetProperty()
            prop.SetSpecular(0.3)
            prop.SetSpecularPower(20)
            prop.SetDiffuse(0.7)
            prop.SetAmbient(0.1)
            prop.SetColor(*SEGMENT_COLORS[i])
            prop.SetOpacity(0.5)
            self._renderer.AddActor(actor)
            self._liver_segment_actors[segment_name] = actor
            i += 1
        picker = LiverSegmentPicker(
            self._renderer, 
            self._interactor, 
            self._liver_segment_actors, 
            self._liver_volumes, 
            self._selected_volume_actor,
        )
        self._renderer.SetBackground(0.1, 0.1, 0.12)
        self._renderer.ResetCamera()
        self._render_window.Render()
        self._interactor.Start()


import os
from mosamatic2.core.tasks.task import Task
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.data.multidicomimage import MultiDicomImage
from mosamatic2.core.utils import (
    convert_numpy_array_to_png_image, 
    convert_muscle_mask_to_myosteatosis_map,
    get_pixels_from_dicom_object,
    AlbertaColorMap, 
    load_numpy_array,
)

LOG = LogManager()


class CreatePngsFromSegmentationsTask(Task):
    INPUTS = ['images', 'segmentations']
    PARAMS = [
        'fig_width', 
        'fig_height',
        'hu_low',
        'hu_high',
        'alpha',
    ]

    def __init__(self, inputs, params, output, overwrite=True):
        super(CreatePngsFromSegmentationsTask, self).__init__(inputs, params, output, overwrite)

    def load_images(self):
        image_data = MultiDicomImage()
        image_data.set_path(self.input('images'))
        if image_data.load():
            return image_data
        raise RuntimeError('Could not load images')

    def load_segmentations(self):
        segmentations = []
        for f in os.listdir(self.input('segmentations')):
            f_path = os.path.join(self.input('segmentations'), f)
            if f.endswith('.seg.npy'):
                segmentations.append(f_path)
        if len(segmentations) == 0:
            raise RuntimeError('Input directory has no segmentation files')
        return segmentations
    
    def collect_img_seg_pairs(self, images, segmentations, file_type='npy'):
        file_type = '.tag' if file_type == 'tag' else '.seg.npy'
        img_seg_pairs = []
        for image in images:
            f_img_path = image.path()
            f_img_name = os.path.split(f_img_path)[1]
            for f_seg_path in segmentations:
                f_seg_name = os.path.split(f_seg_path)[1]
                if file_type == '.seg.npy':
                    f_seg_name = f_seg_name.removesuffix(file_type)
                    if f_seg_name == f_img_name:
                        img_seg_pairs.append((image, f_seg_path))
                elif file_type == '.tag':
                    f_seg_name = f_seg_name.removesuffix(file_type).removesuffix('.dcm')
                    f_img_name = f_img_name.removesuffix('.dcm')
                    if f_seg_name == f_img_name:
                        img_seg_pairs.append((image, f_seg_path))
                else:
                    raise RuntimeError('Unknown file type')
        return img_seg_pairs

    def run(self):
        images = self.load_images()
        segmentations = self.load_segmentations()
        img_seg_pairs = self.collect_img_seg_pairs(images.images(), segmentations, 'npy')
        # nr_steps = len(segmentations)
        nr_steps = len(img_seg_pairs)
        for step in range(nr_steps):
            segmentation = img_seg_pairs[step][1]
            segmentation_name = os.path.split(segmentation)[1]
            segmentation_image = load_numpy_array(segmentation)
            if segmentation_image is not None:
                png_file_name = segmentation_name + '.png'
                convert_numpy_array_to_png_image(
                    segmentation_image, 
                    self.output(),
                    AlbertaColorMap(), 
                    png_file_name,
                    fig_width=int(self.param('fig_width')),
                    fig_height=int(self.param('fig_height')),
                )
                image = img_seg_pairs[step][0]
                image = get_pixels_from_dicom_object(image.object(), normalize=True)
                png_file_name = segmentation_name + '_myosteatosis.png'
                convert_muscle_mask_to_myosteatosis_map(
                    image,
                    segmentation_image,
                    self.output(),
                    png_file_name,
                    hu_low=self.param('hu_low'),
                    hu_high=self.param('hu_high'),
                    alpha=self.param('alpha'),
                )
            else:
                LOG.warning(f'File {segmentation} is not a valid NumPy array file')
            self.set_progress(step, nr_steps)



import os
from dataclasses import dataclass

from PySide6.QtCore import Qt, Signal, QSize
from PySide6.QtGui import QPixmap
from PySide6.QtWidgets import (
    QApplication, QWidget, QLabel, QCheckBox, QVBoxLayout, QGridLayout,
    QScrollArea, QMainWindow, QFileDialog, QPushButton, QHBoxLayout
)


@dataclass(frozen=True)
class SliceItem:
    path: str


class SliceTile(QWidget):
    toggled = Signal(str, bool)  # path, checked

    def __init__(self, item: SliceItem, display_size: QSize, parent=None):
        super().__init__(parent)
        self.item = item
        self.display_size = display_size

        self.image_label = QLabel()
        self.image_label.setAlignment(Qt.AlignCenter)
        self.image_label.setFixedSize(display_size)
        self.image_label.setStyleSheet("QLabel { background: #222; border: 1px solid #444; }")

        dcm_name = os.path.basename(item.path)[:-13] + '.dcm'
        self.checkbox = QCheckBox(dcm_name)
        self.checkbox.setChecked(True)
        self.checkbox.toggled.connect(self._on_toggled)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(6, 6, 6, 6)
        layout.setSpacing(6)
        layout.addWidget(self.image_label)
        layout.addWidget(self.checkbox)

        self._load_and_scale()

    def _load_and_scale(self):
        pix = QPixmap(self.item.path)
        if pix.isNull():
            self.image_label.setText("Failed to load")
            return

        scaled = pix.scaled(
            self.display_size,
            Qt.KeepAspectRatio,
            Qt.SmoothTransformation
        )
        self.image_label.setPixmap(scaled)

    def _on_toggled(self, checked: bool):
        self.toggled.emit(self.item.path, checked)

    def is_checked(self) -> bool:
        return self.checkbox.isChecked()

    def set_checked(self, checked: bool):
        self.checkbox.setChecked(checked)


import os
import math
import tempfile
import shutil
import nibabel as nib
import numpy as np
import SimpleITK as sitk
import matplotlib.pyplot as plt
from totalsegmentator.python_api import totalsegmentator
from mosamatic2.core.tasks.task import Task
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.utils import load_dicom

LOG = LogManager()

TOTAL_SEGMENTATOR_OUTPUT_DIR = os.path.join(tempfile.gettempdir(), 'total_segmentator_output')
TOTAL_SEGMENTATOR_TASK = 'total'
Z_DELTA_OFFSETS = {
    'vertebrae_L3': 0.333,
    'vertebrae_T4': 0.5,
}


class SelectSliceFromScansTask(Task):
    INPUTS = ['scans']
    PARAMS = ['vertebra']

    def __init__(self, inputs, params, output, overwrite):
        super(SelectSliceFromScansTask, self).__init__(inputs, params, output, overwrite)
        self._error_dir = os.path.split(self.output())[0]
        self._error_dir = os.path.join(self._error_dir, 'selectslicefromscanstask_errors')
        os.makedirs(self._error_dir, exist_ok=True)
        self._error_file = os.path.join(self._error_dir, 'errors.txt')
        with open(self._error_file, 'w') as f:
            f.write('Errors:\n\n')
        LOG.info(f'Error directory: {self._error_dir}')

    def write_error(self, message):
        LOG.error(message)
        with open(self._error_file, 'a') as f:
            f.write(message + '\n')

    def load_scan_dirs(self):
        scan_dirs = []
        for d in os.listdir(self.input('scans')):
            scan_dir = os.path.join(self.input('scans'), d)
            if os.path.isdir(scan_dir):
                scan_dirs.append(scan_dir)
        return scan_dirs
    
    def read_ct_series_sitk(self, scan_dir):
        reader = sitk.ImageSeriesReader()
        series_ids = reader.GetGDCMSeriesIDs(scan_dir)
        if not series_ids:
            raise ValueError(f'No DICOM series found in {scan_dir}')
        file_names = reader.GetGDCMSeriesFileNames(scan_dir, series_ids[0])
        reader.SetFileNames(file_names)
        img = reader.Execute()
        return img
    
    def resample_to_reference(self, moving: sitk.Image, reference: sitk.Image, is_label: bool) -> sitk.Image:
        interp = sitk.sitkNearestNeighbor if is_label else sitk.sitkLinear
        return sitk.Resample(
            moving,
            reference,
            sitk.Transform(),
            interp,
            0,  # default value
            moving.GetPixelID()
        )
    
    def centroid_x_index_from_mask(self, mask_ref: sitk.Image) -> int:
        # mask_ref must already be on the CT grid
        arr = sitk.GetArrayFromImage(mask_ref)  # shape: [z, y, x]
        idx = np.argwhere(arr > 0)
        if idx.size == 0:
            # fallback to true mid-sagittal
            size_x = mask_ref.GetSize()[0]
            return size_x // 2
        x_mean = int(round(idx[:, 2].mean()))
        return x_mean
    
    def z_index_from_physical_z(self, ct_img: sitk.Image, z_phys: float, x_index: int) -> int:
        # pick a y roughly mid (or better: y centroid of mask if you want)
        size = ct_img.GetSize()  # (x, y, z)
        y_index = size[1] // 2

        # Convert chosen (x_index, y_index, any z_index) -> physical, then replace z
        p = ct_img.TransformIndexToPhysicalPoint((x_index, y_index, size[2] // 2))
        phys_point = (p[0], p[1], z_phys)

        try:
            ix, iy, iz = ct_img.TransformPhysicalPointToIndex(phys_point)
        except RuntimeError:
            # if z_phys is slightly out of range, clamp later
            iz = None

        if iz is None:
            # clamp using physical bounds
            # safest crude clamp: convert all slice centers to physical z and find closest
            z_centers = []
            for k in range(size[2]):
                pk = ct_img.TransformIndexToPhysicalPoint((x_index, y_index, k))
                z_centers.append(pk[2])
            z_centers = np.array(z_centers)
            iz = int(np.argmin(np.abs(z_centers - z_phys)))

        iz = int(np.clip(iz, 0, size[2]-1))
        return iz
    
    def plot_sagittal_with_vertebra_overlay(self, scan_dir, mask_file, z_vert_phys_mm, out_png):
        ct = self.read_ct_series_sitk(scan_dir)
        vert_mask = sitk.ReadImage(mask_file)
        vert_mask_ref = self.resample_to_reference(vert_mask, ct, is_label=True)
        x_idx = self.centroid_x_index_from_mask(vert_mask_ref)
        z_idx = self.z_index_from_physical_z(ct, z_vert_phys_mm, x_idx)
        ct_arr = sitk.GetArrayFromImage(ct).astype(np.float32)
        mk_arr = sitk.GetArrayFromImage(vert_mask_ref).astype(np.uint8)
        sag_ct = ct_arr[:, :, x_idx]
        sag_mk = mk_arr[:, :, x_idx]
        vmin, vmax = np.percentile(sag_ct, (1, 99))
        sy = ct.GetSpacing()[1]  # y spacing (mm)
        sz = ct.GetSpacing()[2]  # z spacing (mm)
        aspect = sz / sy
        plt.figure(figsize=(7, 9))
        plt.imshow(sag_ct, cmap="gray", vmin=vmin, vmax=vmax, origin="lower", aspect=aspect)
        plt.imshow(sag_mk, alpha=0.35, origin="lower", aspect=aspect)
        plt.axhline(z_idx, linewidth=2)  # line across the vertebra axial slice position
        plt.title("Sagittal view with vertebral mask overlay + selected vertebra axial slice")
        plt.axis("off")
        plt.savefig(out_png, bbox_inches="tight", dpi=200)

    def extract_masks(self, scan_dir):
        os.makedirs(TOTAL_SEGMENTATOR_OUTPUT_DIR, exist_ok=True)
        totalsegmentator(input=scan_dir, output=TOTAL_SEGMENTATOR_OUTPUT_DIR, fast=True)
        if not os.path.isfile(os.path.join(TOTAL_SEGMENTATOR_OUTPUT_DIR, 'vertebrae_L3.nii.gz')):
            raise Exception(f'{scan_dir}: vertebrae_L3.nii.gz does not exist')
        # os.system(f'TotalSegmentator -i {scan_dir} -o {TOTAL_SEGMENTATOR_OUTPUT_DIR} --fast')

    def delete_total_segmentator_output(self):
        if os.path.exists(TOTAL_SEGMENTATOR_OUTPUT_DIR):
            shutil.rmtree(TOTAL_SEGMENTATOR_OUTPUT_DIR)

    def get_z_delta_offset_for_mask(self, mask_name):
        if mask_name not in Z_DELTA_OFFSETS.keys():
            return None
        return Z_DELTA_OFFSETS[mask_name]

    def find_slice(self, scan_dir, vertebra):
        if vertebra == 'L3':
            vertebral_level = 'vertebrae_L3'
        elif vertebra == 'T4':
            vertebral_level = 'vertebrae_T4'
        else:
            self.write_error(f'{scan_dir}: Unknown vertbra {vertebra}. Options are "L3" and "T4"')
            return None
        # Find Z-positions DICOM images
        z_positions = {}
        for f in os.listdir(scan_dir):
            f_path = os.path.join(scan_dir, f)
            try:
                p = load_dicom(f_path, stop_before_pixels=True)
                if p is not None and hasattr(p, "ImagePositionPatient"):
                    z_positions[p.ImagePositionPatient[2]] = f_path
            except Exception as e:
                self.write_error(f"{scan_dir}: Failed to load DICOM {f_path}: {e}")
                break
        if not z_positions:
            self.write_error(f"{scan_dir}: No valid DICOM z-positions found.")
            return None
        # Find Z-position vertebral image
        mask_file = os.path.join(TOTAL_SEGMENTATOR_OUTPUT_DIR, f'{vertebral_level}.nii.gz')
        if not os.path.exists(mask_file):
            self.write_error(f"{scan_dir}: Mask file not found: {mask_file}")
            return None
        try:
            mask_obj = nib.load(mask_file)
            mask = mask_obj.get_fdata()
            affine_transform = mask_obj.affine
        except Exception as e:
            self.write_error(f"{scan_dir}: Failed to load mask {mask_file}: {e}")
            return None
        indexes = np.array(np.where(mask == 1))
        if indexes.size == 0:
            self.write_error(f"{scan_dir}: No voxels found in mask {mask_file} for {vertebral_level}")
            return None        
        try:
            index_min = indexes.min(axis=1)
            index_max = indexes.max(axis=1)
        except ValueError as e:
            self.write_error(f"{scan_dir}: Invalid indexes array for {vertebral_level}: {e}")
            return None
        world_min = nib.affines.apply_affine(affine_transform, index_min)
        world_max = nib.affines.apply_affine(affine_transform, index_max)
        z_direction = affine_transform[:3, 2][2]
        if z_direction == 0:
            self.write_error(f"{scan_dir}: Affine z-direction is zero.")
            return None
        z_sign = math.copysign(1, z_direction)
        z_delta_offset = self.get_z_delta_offset_for_mask(vertebral_level)
        if z_delta_offset is None:
            return None
        z_delta = 0.333 * abs(world_max[2] - world_min[2]) # This needs to be vertebra-specific perhaps
        z_l3 = world_max[2] - z_sign * z_delta
        # Find closest L3 image in DICOM set
        positions = sorted(z_positions.keys())
        closest_file = None
        for z1, z2 in zip(positions[:-1], positions[1:]):
            if min(z1, z2) <= z_l3 <= max(z1, z2):
                closest_z = min(z_positions.keys(), key=lambda z: abs(z - z_l3))
                closest_file = z_positions[closest_z]
                LOG.info(f'Closest image: {closest_file}')
                break
        if closest_file is None:
            self.write_error(f"{scan_dir}: No matching slice found.")
        return closest_file, z_l3
    
    def run(self):
        scan_dirs = self.load_scan_dirs()
        vertebra = self.param('vertebra')
        nr_steps = len(scan_dirs)
        for step in range(nr_steps):
            scan_dir = scan_dirs[step]
            scan_name = os.path.split(scan_dir)[1]
            errors = False
            LOG.info(f'Processing {scan_dir}...')
            try:
                self.extract_masks(scan_dir)
            except Exception as e:
                self.write_error(f'{scan_dir}: Could not extract masks [{str(e)}]. Skipping scan...')
                errors = True
            if not errors:
                file_path, z_vert = self.find_slice(scan_dir, vertebra)
                if file_path is not None:
                    extension = '' if file_path.endswith('.dcm') else '.dcm'
                    target_file_path = os.path.join(self.output(), vertebra + '_' + scan_name + extension)
                    shutil.copyfile(file_path, target_file_path)
                    mask_file = os.path.join(TOTAL_SEGMENTATOR_OUTPUT_DIR, f'vertebrae_{vertebra}.nii.gz')
                    output_png = os.path.join(self.output(), f"{vertebra}_{scan_name}_sagittal.png")
                    self.plot_sagittal_with_vertebra_overlay(scan_dir, mask_file, z_vert, output_png)
                else:
                    self.write_error(f'{scan_dir}: Could not find slice for vertebral level: {vertebra}')
                    errors = True
                self.delete_total_segmentator_output()
            if errors:
                LOG.info(f'Copying problematic scan {scan_dir} to error directory: {self._error_dir}')
                scan_error_dir = os.path.join(self._error_dir, scan_name)
                shutil.copytree(scan_dir, scan_error_dir)
            self.set_progress(step, nr_steps)